(ns lu-preprocessing.stat)


;;; private methods

(defn- get-positions
  "Returns a list of positions.
    rows: A hash-map with row keys and corresponding columns.
   returns: A list of positions (row-column pairs)."
  [rows]
  (->> ((juxt #(map (fn[[v1 v2]]
                      (-> (count v2)
                          (repeat v1)))
                    %)
              vals)
        rows)
       (map #(apply concat %))
       (apply map vector)))

(defn- reverse-map
  "Returns a reversed map with position keys.
      source-map: A map with value keys.
   returns: A reversed hash-map with position keys."
  [source-map]
  (->> source-map
       (map (fn[[value rows]]
              (->> (get-positions rows)
                   (map vector
                        (repeat value))
                   (map reverse)
                   (apply concat))))
       (apply concat)
       (apply hash-map)))


;;; public methods

(defn get-neighbours
  "Returns a list of neighbours for groupped cells.
      values: A hash-map of group ids with their positions (by row).
   returns: A hash-map with cell group identifiers and its neighbours.

   ## Usage

      (require '[lu-preprocessing.stat :refer :all])

      (get-neighbours {0 {0 [0 1 2 4] 1 [4 5] 8 [5]}
                       1 {0 [5]}
                       2 {1 [0 1 2]}
                       3 {1 [3] 7 [0 1 2 3 4] 8 [0 1 2 3 4]}
                       4 {3 [1]}
                       5 {5 [0 1 2 3 4 5] 6 [0 1 2 3 4 5] 7 [5]}
                       6 {9 [0]}})
      => {0 [1 2 3 5]
          1 [0]
          2 [0 3]
          3 [0 2 5 6]
          4 []
          5 [0 3]
          6 [3]}
  "
  [values]
  (let [[n-rows n-cols]
        (->> (vals values)
             ((juxt #(->> (map keys %)
                          (apply concat)
                          (apply max))
                    #(->> (map vals %)
                          flatten
                          (apply max)))))
        vals-by-positions
        (reverse-map values)]
    (->> (pmap (fn[[value positions]]
                (let [get-row-pos
                      (fn[row1 row2]
                        (if (or (neg? row1)
                                (> row1 n-rows))
                          {}
                          (->> (get positions row2)
                               (map (juxt dec identity inc))
                               (apply concat)
                               distinct
                               (filter #(not (or (neg? %)
                                                 (> % n-cols))))
                               vec
                               (hash-map row1))))]
                  (->> (map (juxt #(get-row-pos (dec %) %)
                                  #(get-row-pos % %)
                                  #(get-row-pos (inc %) %))
                            (keys positions))
                       (map #(apply merge %))
                       (apply merge-with
                              #(->> (apply concat %&)
                                    distinct
                                    vec))
                       (hash-map value)
                       reverse-map
                       keys
                       (map #(get vals-by-positions %))
                       distinct
                       (filter #(not (or (nil? %)
                                         (= value %))))
                       sort
                       vec)))
              values)
         (zipmap (keys values)))))
