(ns data-cleaning.core
  (:require [data-cleaning.land-use-sample :as land-use-sample]
            [utilities.writer :as writer])
  (:gen-class))


;;; private defs

;; Defines a path to bin folder.
(def ^:private target-folder "bin")


;;; private methods

(defn- print-message
  "Prints a multiline message.
      message: Validation message."
  [message]
  (doall (map println
              message)))

(defn -main
  "Application entry point.
      settings-path: Path to the settings file.

   Steps include: 1) filtering of cells,
                  2) transformation of the land use map
                     by counting the number of cells with urban land use
                     in the cell neighbourhood,
                  3) transformation of other maps with log and rescaling to [0,1],
                     if specified.

   In case of valid original maps,
   saves to disk
      a) cleaned .asc files,
            output: bin/[\"land-use.asc\"]
                    bin/[\"map1.asc\"]
                    bin/[\"map1.asc\"]
                    ...
      b) a sorted list of Moore neighbourhoods by their frequency in the land use map
         (in csv and latex formats)
            output: bin/[\"land-use.asc\"]-neighbourhoods.csv
                    bin/[\"land-use.asc\"]-neighbourhoods.tex
   prints to terminal
      c) a list of validation warnings.

   Otherwise, prints to terminal a list of validation errors.

   ## Usage

       $ lein run \"settings.xml\"
  "
  [settings-path & args]
  (let [sample (land-use-sample/clean settings-path)]
    (if (not (nil? (get sample :error)))
      (do
        (print-message (get sample :error))
        (java.lang.System/exit 1))
      (do
        (doall
         (map #(print-message %)
              (get sample :warnings)))
        (time
         (do
           (doall
            (map #(writer/txt-file target-folder
                                   %
                                   (get-in sample
                                           [:maps %]))
                 (keys (get sample :maps))))
           (doall
            (map (fn [[path content]]
                   (writer/csv-file target-folder
                                    path
                                    content))
                 (get-in sample [:statistics :csv])))
           (doall
            (map (fn [[path content]]
                   (writer/txt-file target-folder
                                    path
                                    content))
                 (get-in sample [:statistics :text])))))))))
