(ns data-cleaning.land-use-sample
  (:require [data-cleaning.settings :as settings]
            [data-cleaning.map :as map]
            [data-cleaning.validation :as validation]
            [data-cleaning.output-provider :as provider]
            [data-cleaning.sample :as sample]
            [data-cleaning.sampling :as sampling]))


;;; private methods

(defn- read-properties
  "Returns a collection of the raster file preambles.
      settings: Settings object.
   returns: Collection of the raster file preambles."
  [settings]
  (-> (keys (:other settings))
      (conj (:land-use settings))
      ((juxt identity
             (partial map
                      map/read-properties)))
      (#(apply zipmap %))))

(defn- unite-na-in-row
  "Returns a union of na and mask value positions in the maps for the specified row.
      map-properties-list: Collection of .asc preamble data per file.
      mask-positions: Positions of mask values.
      row: Row number.
   returns: A union of na and mask value positions."
  [map-properties-list
   mask-positions
   row]
  (->> map-properties-list
       (map #(get-in % [:na-cells row]))
       (cons (get mask-positions row))
       flatten
       distinct
       (filter (complement nil?))
       sort
       vec))

(defn- form-preamble-data
  "Returns an .asc preamble data with a union of na and mask value positions.
      map-properties-per-file: Collection of .asc preamble data per file.
      mask-positions: Positions of mask values.
   returns: An .asc preamble data."
  [map-properties-per-file
   mask-positions]
  (->> (vals map-properties-per-file)
       (#(assoc (select-keys (first %)
                             [:declared-size
                              :left-bottom-corner
                              :cell-size
                              :cell_type])
                :na-cells
                (reduce (fn [na-positions row]
                          (let [positions-in-row
                                (unite-na-in-row %
                                                 mask-positions
                                                 row)]
                            (if-not (empty? positions-in-row)
                              (assoc na-positions
                                     row
                                     positions-in-row)
                              na-positions)))
                        {}
                        ((comp range
                               first
                               :declared-size
                               first)
                         %))))))

(defn- get-mask-positions
  "Returns a list of mask value positions.
      settings: Settings object.
   returns: A list of mask value positions."
  [settings]
  (-> (:mask settings)
      (#(if (nil? %)
         []
         (map/read-positions %)))))

(defn- create-stat-file
  "Returns a file name and content of the summary statistics file.
      get-content: Method to get file content.
      file-key: A key of file group.
      land-use-name: Land use file name.
      neighbourhood-data: Data about different types of Moore neighbourhood.
   returns: A file name and content of the summary statistics file."
  [get-content
   file-key
   land-use-name
   neighbourhood-data]
  (->> (get-content neighbourhood-data)
       (hash-map (str land-use-name
                      "-neighbourhoods"
                      (if (= :csv
                             file-key)
                        ".csv"
                        ".tex")))
       (hash-map file-key)))

(defn- get-statistics
  "Returns a collection of the summary statistics files.
      neighbour-matrices: Matrices with neighbour land-use types.
      land-use-name: Land use file name.
   returns: A hash-map of the summary statistics files."
  [neighbour-matrices
   land-use-name]
  (->> (sample/get-neighbourhood-data neighbour-matrices)
       ((juxt (partial create-stat-file
                       provider/stat-to-csv
                       :csv
                       land-use-name)
              (partial create-stat-file
                       provider/stat-to-latex-tabular
                       :text
                       land-use-name)))
       (into {})))

(defn- get-maps
  "Returns a collection of the map raster files.
      matrices: A hash map of map matrices and file names.
      map-properties: .asc preamble data.
   returns: A hash-map of map raster files."
  [matrices map-properties]
  (->> (:var-matrices matrices)
       (#(apply assoc
                %
                (:land-use-matrix matrices)))
       ((juxt keys
              #(->> (vals %)
                    (map (fn[matrix]
                           (provider/map-to-raster-file matrix
                                                        map-properties))))))
       (apply zipmap)))


;;; public methods

(defn clean
  "Forms a sample of the cleaned .asc map files.

      settings-path: A path to the xml settings file.


  Steps include: 1) filtering of cells,
                 2) transformation of the land use map
                    by counting the number of cells with urban land use
                    in the cell neighbourhood,
                 3) transformation of other maps with log and rescaling to [0,1],
                    if specified.

  1) Excludes cells from every map if
         - a cell value is not defined originally
           (in any of the original maps specified),
         - any value in the cell Moore neighbourhood is not defined originally
           (in any of the original maps specified),
         - a cell belongs to the map-matrix border,
         - a cell is masked in the \"black-list\" file,
         - any value in the cell Moore neighbourhood is masked in the \"black-list\" file.
     \"Black-list\" file should mask cells with \"1\" in order to exclude these cells
     from the resultant maps.

  2) Original land use map should contain 0-1 cell values, where 0 is a non-urban land
     (e.g., vegetation, wetlands, agricultural land and water) and 1 is an urban land
     (e.g., all artificial surfaces).

     Every cell value in the land use map is substituted with the number of cells with
     urban land use in the cell Moore neighbourhood.

  3) If specified, transforms cell values with ln function or rescaling to the [0,1] range.
     Rescaling to the [0,1] range is done using the following transformation:
        x'=(x-x_min)/(x_max-x_min),
     where x is an original value. The minimum and maximum values (x_min, x_max respectively)
     are taken from the original cell values.

     Logarithmic transformation should be specified with 'log' attribute in the settings file.
     Rescaling to [0,1] transformation should be specified with 'unit-rescaling' attribute
     in the settings file.


  In case of valid original maps,
  returns a hash-map of
     a) cleaned .asc files content with file names,
     b) a sorted list of Moore neighbourhoods by their frequency in the land use map
        (in csv and latex formats) with names [\"land-use.asc\"-neighbourhoods.csv] and
        [\"land-use.asc\"-neighbourhoods.tex],
     c) a list of validation warnings.

  Otherwise, returns a list of validation errors.

  ## Usage

    (require '[data-cleaning.land-use-sample :refer :all])

    (clean \"settings.xml\")

  "
  [settings-path]
  (let [settings
        (settings/read-settings settings-path)
        map-properties-per-file
        (read-properties settings)]
    (->> (validation/validate-maps map-properties-per-file)
         ((juxt #(-> (select-keys % [:warnings])
                     (update :warnings
                             (fn[w]
                               (reduce (fn[warnings next-warning]
                                         (->> (provider/convert-to-warning-message next-warning)
                                              (conj warnings)))
                                       []
                                       w))))
                #(->> (:error %)
                      ((fn[error]
                         (if-not (nil? error)
                           (->> (provider/convert-to-error-message error)
                                (hash-map :error))
                           (let [map-properties
                                 (->> (get-mask-positions settings)
                                      (form-preamble-data map-properties-per-file))]
                             (->> (sampling/clean-and-transform map-properties
                                                                settings)
                                  ((juxt (fn[matrices]
                                           (get-maps matrices map-properties))
                                         (fn[matrices]
                                           (->> ((juxt :neighbour-matrices
                                                       (comp first
                                                             :land-use-matrix))
                                                 matrices)
                                                (apply get-statistics)))))
                                  (zipmap [:maps
                                           :statistics])))))))))
         (apply merge))))
