(ns data-cleaning.settings
 (:require [utilities.reader :as reader]))


;;; private defs

;; Defines a group attribute value for land use map.
(def ^:private lu-map "land-use")

;; Defines a group attribute value for mask map.
(def ^:private mask-map "mask")

;; Defines a group attribute value for region map.
(def ^:private region-map "region")


;;; private methods

(defn- get-paths
  "Returns file paths from xml attributes
   for the selected raster group.
      files: Collection of objects with xml attributes.
      group-name: Selected group of the raster files
                  in the dataset.
   returns: File paths for the selected raster group."
  [files group-name]
  (->> files
       (filter #(= (:group %) group-name))
       (map #(:path %))
       vec))


;;; public methods

(defn read-settings
  "Reads dataset settings file.
      settings-path: Path to the settings file.
   returns: File paths and relevant information
            for the dataset.

  ## Usage

      (require '[data-cleaning.settings :refer :all])

      ;; \"settings.xml\"

      ;; <?xml version=\"1.0\"?>
      ;; <files>
      ;; <raster path = \"land-use-file\"
      ;;         group = \"land-use\" />
      ;; <raster path = \"economic-indicator-file\" />
      ;; <raster path = \"economy1-file\" />
      ;; <raster path = \"economy2-file\"
      ;;         transform = \"log\" />
      ;; <raster path = \"zoning-file\"
      ;;         group = \"mask\" />
      ;; </files>


      (read-settings \"settings.xml\")
      => {:land-use \"land-use-file\"
          :other {\"economic-indicator-file\" :identity
                  \"economy1-file\" :identity
                  \"economy2-file\" :log}
          :mask \"zoning-file\"}
  "
  [settings-path]
  (->> (reader/read-xml-file settings-path)
       (:content)
       (filter #(= (:tag %) :raster))
       (map #(:attrs %))
       ((fn[xml-elements]
          (hash-map :land-use
                   (first (get-paths xml-elements
                                     lu-map))
                   :mask
                   (first (get-paths xml-elements
                                     mask-map))
                   :other
                   (->> (filter #(->> (:group %)
                                      ((juxt (partial = lu-map)
                                             (partial = mask-map)))
                                      (every? false?))
                                xml-elements)
                        (map #(list (:path %) (:transform %)))
                        (apply map vector)
                        (#(zipmap (first %)
                                  (->> (second %)
                                       (map (fnil keyword
                                                  :identity)))))))))))

(defn read-region
  "Reads paths to the region maps from the settings file.
      settings-path: Path to the settings file.
   returns: Paths to the region maps.

  ## Usage

      (require '[data-cleaning.settings :refer :all])

      ;; \"settings.xml\"

      ;; <?xml version=\"1.0\"?>
      ;; <files>
      ;; <raster path = \"land-use-file\"
      ;;         group = \"land-use\" />
      ;; <raster path = \"economic-indicator-file\" />
      ;; <raster path = \"economy1-file\" />
      ;; <raster path = \"economy2-file\"
      ;;         transform = \"log\" />
      ;; <raster path = \"zoning-file\"
      ;;         group = \"mask\" />
      ;; </files>


      (read-region \"settings.xml\")
      => \"region-file\"
  "
  [settings-path]
  (->> (reader/read-xml-file settings-path)
       (:content)
       (filter #(= (:tag %) :raster))
       (map #(:attrs %))
       (filter #(= (:group %) region-map))
       (map #(:path %))))
