(ns data-cleaning.validation)


;;; private methods

(defn- apply-validation
  "Applies a validation rule.
      validate: Method to apply validation rule.
      error-message: Validation error message.
      get-trace: Method to get trace in case of validation failure.
   returns: nil, if validation was passed successfully;
            otherwise, validation message and trace."
  [validate error-message get-trace]
  (if (validate)
    (identity nil)
    {:message error-message
     :trace (get-trace)}))

(defn- declared-map-size-valid?
  "Indicates whether map structure is valid for the declared map size.
      map-properties: Map properties.
   returns: False, if number of rows doesn't equal declared one,
            or number of columns in each rows doesn't equal declared one;
            otherwise, true."
  [map-properties]
  (let [size (:declared-size map-properties)
        rows (:rows map-properties)]
    (and (= (first size)
            (count rows))
         (every? #(= (first rows) %)
                 rows)
         (= (second size)
            (first rows)))))

(defn- validate-declared-map-sizes
  "Validates whether declared map sizes are valid.
      map-properties-per-file: Collection of .asc preamble data per file.
   returns: Validation error, if declared map sizes are invalid;
            otherwise, nil."
  [map-properties-per-file]
  (let [invalid-maps
        (filter #(not (declared-map-size-valid? (second %)))
                map-properties-per-file)]
    (apply-validation #(empty? invalid-maps)
                      "Declared number of rows and columns doesn't fit data."
                      #(zipmap (map first invalid-maps)
                               (map (fn[[file map-properties]]
                                      (let [rows (:rows map-properties)]
                                        {:rows (count rows)
                                         :columns rows}))
                                    invalid-maps)))))

(defn- validate-equality
  "Validates whether selected properties in the maps are identical.
      map-properties-per-file: Collection of .asc preamble data per file.
      property-keys: Keys of selected properties.
      error-message: Validation error message.
      get-trace-record: Method to get trace record for the map file.
    returns: Vaildation error, if selected properties in the maps
             are not identical;
             otherwise, nil."
  [map-properties-per-file
   property-keys
   error-message
   get-trace-record]
  (let [properties-per-file
        (map #(select-keys (second %) property-keys)
             map-properties-per-file)]
    (apply-validation  #(every? (fn[properties]
                                  (= properties (first properties-per-file)))
                                properties-per-file)
                       error-message
                       #(into {}
                              (map (fn[[file map-properties]]
                                     [file (get-trace-record map-properties)])
                                   map-properties-per-file)))))


(defn- validate-general-properties
  "Validates whether map general properties are identical.
      map-properties-per-file: Collection of .asc preamble data per file.
   returns: Validation error with general properties per file,
            if they are not identical;
            otherwise, nil."
  [map-properties-per-file]
  (validate-equality map-properties-per-file
                     [:declared-size
                      :cell-size
                      :left-bottom-corner]
                     "Grid declaration is different in the source maps."
                     (fn[map-properties]
                       (let [size (:declared-size map-properties)]
                         (conj
                          (hash-map :rows (first size)
                                    :columns (second size))
                          (into {}
                                (select-keys map-properties
                                             [:cell-size
                                              :left-bottom-corner])))))))

(defn- validate-na-values
  "Validates whether na value positions are identical in maps.
      map-properties-per-file: Collection of .asc preamble data per file.
   returns: Validation error with na value count,
            if their positions are not identical;
            otherwise, nil."
  [map-properties-per-file]
  (validate-equality map-properties-per-file
                     [:na-cells]
                     "NA value positions are different in the source maps."
                     (fn[map-properties]
                       (reduce (fn [seed row]
                                 (+ seed (count (second row))))
                               0
                               (:na-cells map-properties)))))

(defn- validate-successively
  "Applies validation methods successively.
   Interupts validation if some validation method was not passed.
      map-properties-per-file: Collection of .asc preamble data per file.
      validation-methods: Collection of validation methods.
   returns: Validation error, if some validation method was not passed;
            otherwise, nil."
  [map-properties-per-file & validation-methods]
  (let [error
        ((first validation-methods) map-properties-per-file)
        rest-methods
        (next validation-methods)]
    (if (or (not (nil? error))
            (nil? rest-methods))
      (identity error)
      (recur map-properties-per-file
             rest-methods))))

(defn- validate-warnings
  "Applies validation methods to get warnings.
      map-properties-per-file: Collection of .asc preamble data per file.
      error: Validation error.
   returns: Empty list, if validation error was encountered
            or no warnings were returned;
            otherwise, validation warnings."
  [map-properties-per-file
   error]
  (if-not (nil? error)
    []
    (reduce (fn[warnings method]
              (if-let [warning (method map-properties-per-file)]
                (conj warnings warning)
                warnings))
            []
            (list validate-na-values))))


;;; public methods

(defn validate-maps
  "Validates .asc preamble data.
      map-properties-per-file: Collection of .asc preamble data per file.
   returns: Validation errors, if maps are inconsistent.
            Otherwise, nil.

  ## Usage

      (require '[data-cleaning.validation :refer :all])

      (results (validate-maps {\"1.asc\"
                               {:declared-size [1 1]
                                :rows [1 1]
                                :cell-size 1.0
                                :left-bottom-corner [1.0 1.0]
                                :na-cells {}}
                               \"2.asc\"
                               {:declared-size [1 1]
                                :rows [1]
                                :cell-size 1.0
                                :left-bottom-corner [1.0 1.0]
                                :na-cells {}}
                               \"3.asc\"
                               {:declared-size [3 2]
                                :rows [2 1 2]
                                :cell-size 1.0
                                :left-bottom-corner [1.0 1.0]
                                :na-cells {}}}))
      (get results :error)
      => {:message \"Declared number of rows and columns doesn't fit data.\"
          :trace {\"1.asc\" {:rows 2 :columns [1 1]}
                  \"3.asc\" {:rows 3 :columns [2 1 2]}}}
  "
  [map-properties-per-file]
  (let [error (validate-successively map-properties-per-file
                                     validate-declared-map-sizes
                                     validate-general-properties)]
    (hash-map :error error
              :warnings (validate-warnings map-properties-per-file error))))
