(ns data-cleaning.land-use-sample-test
  (:require [clojure.test :refer :all]
            [data-cleaning.land-use-sample :refer :all]))


;;; tests

(deftest include-source-maps-and-neighbour-maps
  (testing "A land use sample should include
    a map with land use type distribution,
    maps with land use type distribution for each cell in the Moore neighbourhood (8 maps),
    maps with explanatory economic variables,
    a map with the dependent economic indicator and
    maps with neighbour cells for the dependent economic indicator."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (= (count (keys (get sample :maps)))
              6)))))

(deftest clean-land-use-map
  (testing "A land use sample should include a filtered map
    with land use type distribution."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (= (get-in sample [:maps "land_use.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 4 -9999 -9999 -9999 -9999"
               "-9999 -9999 5 4 -9999 -9999 -9999"
               "-9999 -9999 5 4 -9999 -9999 -9999"
               "-9999 2 -9999 -9999 -9999 2 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"])))))

(deftest clean-maps-of-variables
  (testing "A land use sample should include filtered maps
    with supplied variables."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (= (get-in sample [:maps "economy1.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 7 -9999 -9999 -9999 -9999"
               "-9999 -9999 12 13 -9999 -9999 -9999"
               "-9999 -9999 18 19 -9999 -9999 -9999"
               "-9999 24 -9999 -9999 -9999 28 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"]))

       (is (= (get-in sample [:maps "economy2.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 0.666666666666667 -9999 -9999 -9999 -9999"
               "-9999 -9999 0.666666666666667 0.666666666666667 -9999 -9999 -9999"
               "-9999 -9999 0.666666666666667 0.666666666666667 -9999 -9999 -9999"
               "-9999 0.666666666666667 -9999 -9999 -9999 0.666666666666667 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"]))

       (is (= (get-in sample [:maps "indicator.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 0.000000000000000 -9999 -9999 -9999 -9999"
               "-9999 -9999 1.098612288668110 1.098612288668110 -9999 -9999 -9999"
               "-9999 -9999 0.000000000000000 1.945910149055313 -9999 -9999 -9999"
               "-9999 1.791759469228055 -9999 -9999 -9999 2.079441541679836 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"]))

       (is (= (get-in sample [:maps "region1.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 0.000000000000000 -9999 -9999 -9999 -9999"
               "-9999 -9999 0.000000000000000 0.000000000000000 -9999 -9999 -9999"
               "-9999 -9999 0.333333333333333 0.500000000000000 -9999 -9999 -9999"
               "-9999 0.166666666666667 -9999 -9999 -9999 0.833333333333333 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"]))

       (is (= (get-in sample [:maps "region3.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 1 -9999 -9999 -9999 -9999"
               "-9999 -9999 1 1 -9999 -9999 -9999"
               "-9999 -9999 1 1 -9999 -9999 -9999"
               "-9999 1 -9999 -9999 -9999 1 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"])))))

(deftest return-error-when-validation-fails
  (testing "Error message should be returned
    if validation of the original maps fails."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings2.xml")]

       ; Assert
       (is (nil? (get sample :maps)))
       (is (= (get sample :error)
              '("Error: Grid declaration is different in the source maps."
               " file test/data_cleaning/test_files/test_case/economy1.asc:\n\t{:cell-size 200.0, :columns 7, :left-bottom-corner [1.0 2.0], :rows 6}"
               " file test/data_cleaning/test_files/test_case/economy2.asc:\n\t{:cell-size 200.0, :columns 7, :left-bottom-corner [1.0 2.0], :rows 6}"
               " file test/data_cleaning/test_files/test_case/invalid_indicator.asc:\n\t{:cell-size 200.0, :columns 5, :left-bottom-corner [1.0 2.0], :rows 5}"
               " file test/data_cleaning/test_files/test_case/land_use.asc:\n\t{:cell-size 200.0, :columns 7, :left-bottom-corner [1.0 2.0], :rows 6}"))))))

(deftest nil-error-if-origin-sample-is-valid
  (testing "Error should be nil if original sample passed validation successfully."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (nil? (get sample :error))))))

(deftest return-warnings-when-validation-indicates
  (testing "Warning message should be returned
    if validation of the original maps returns any."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings5.xml")]

       ; Assert
       (is (= (get sample :warnings)
              [
              '("Warning: NA value positions are different in the source maps."
                " file test/data_cleaning/test_files/test_case/economy1.asc:\n\t9"
                " file test/data_cleaning/test_files/test_case/economy2.asc:\n\t9"
                " file test/data_cleaning/test_files/test_case/indicator.asc:\n\t9"
                " file test/data_cleaning/test_files/test_case/land_use.asc:\n\t9"
                " file test/data_cleaning/test_files/test_case/region1.asc:\n\t6"
                " file test/data_cleaning/test_files/test_case/region2.asc:\n\t42")])))))

(deftest unite-na-value-positions
  (testing "The list of NA value positions should be a union of each map NA list."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings3.xml")]

       ; Assert
       (is (= (get-in sample [:maps "land_use2.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 4 4 2 -9999 -9999"
               "-9999 3 5 4 3 0 -9999"
               "-9999 3 5 4 3 1 -9999"
               "-9999 -9999 -9999 2 2 2 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"])))))

(deftest mask-values-as-na
  (testing "The list of NA value positions should be supplemented with mask value positions."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings4.xml")]

       ; Assert
       (is (= (get-in sample [:maps "land_use2.asc"])
              ["ncols 7"
               "nrows 6"
               "xllcorner 1.0"
               "yllcorner 2.0"
               "cellsize 200.0"
               "NODATA_value -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"
               "-9999 -9999 4 -9999 -9999 -9999 -9999"
               "-9999 3 5 -9999 -9999 -9999 -9999"
               "-9999 3 5 -9999 -9999 -9999 -9999"
               "-9999 -9999 -9999 -9999 2 2 -9999"
               "-9999 -9999 -9999 -9999 -9999 -9999 -9999"])))))

(deftest create-csv-files-with-stat
  (testing "Neighbourhood frequencies and summary stats for the dependent economic variable
            by neighbourhood type should be supplemented with csv file."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (= (get-in sample [:statistics :csv "land_use.asc-neighbourhoods.csv"])
              '(["No."
                 "Left-upper" "Center-upper" "Right-upper"
                 "Left-middle" "Center-middle" "Right-middle"
                 "Left-lower" "Center-lower" "Right-lower"
                 "Rank" "Frequency"]
                ["1" "0" "1" "0" "0" "0" "1" "0" "1" "1" "1-7" "1"]
                ["2" "0" "0" "1" "0" "1" "1" "1" "0" "1" "1-7" "1"]
                ["3" "0" "1" "0" "1" "1" "0" "0" "1" "0" "1-7" "1"]
                ["4" "0" "1" "1" "1" "0" "1" "1" "0" "0" "1-7" "1"]
                ["5" "1" "1" "0" "0" "1" "0" "0" "0" "1" "1-7" "1"]
                ["6" "0" "1" "0" "0" "1" "0" "0" "0" "0" "1-7" "1"]
                ["7" "0" "0" "0" "1" "0" "0" "0" "0" "1" "1-7" "1"]))))))

(deftest create-latex-files-with-stat
  (testing "Neighbourhood frequencies and summary stats for the dependent economic variable
            by neighbourhood type should be supplemented with latex file."
    (
     ; Act
     let [sample (clean "test/data_cleaning/test_files/test_case/settings.xml")]

       ; Assert
       (is (= (get-in sample [:statistics :text "land_use.asc-neighbourhoods.tex"])
              ["1"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "2"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "3"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "4"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "5"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "6"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"
               "7"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1-7 & 1 \\vspace{2mm} \\\\"])))))


;;; test grouping

(deftest clean-test
  (testing "Integration testing - a land use sample should be formed
    from the filtered raster files:\n"
    (include-source-maps-and-neighbour-maps)
    (clean-land-use-map)
    (clean-maps-of-variables)
    (return-error-when-validation-fails)
    (nil-error-if-origin-sample-is-valid)
    (return-warnings-when-validation-indicates)
    (unite-na-value-positions)
    (mask-values-as-na)
    (create-csv-files-with-stat)
    (create-latex-files-with-stat)))

;;; tests in the namespace

(defn test-ns-hook
  "summary: Explicit definition of tests in the namespace.
   returns: Explicit definition of tests in the namespace."
  []
  (clean-test))
