(ns data-cleaning.spatial-maps-test
  (:require [clojure.test :refer :all]
            [data-cleaning.spatial-maps :refer :all]
            [utilities.floating-point-comparison :refer :all]))


;;; tests

(deftest form-binary-layers-for-neighbour-cells-test
  (testing "Binary maps should be formed for neighbour cells
    by the cell type category."
    (
     ; Arrange
     let [land-use-map [nil nil nil nil nil
                        nil "1" 2  nil nil
                        nil "1" "1" nil nil
                        nil "1" "1" nil nil
                        nil 2 2   3   nil
                        nil "1" nil "1" nil
                        nil nil nil nil nil]
          origin [2 2 "1" "1" 2
                  "1" "1" 2 2 nil
                  2 "1" "1" "1" nil
                  3 "1" "1" "1" "1"
                  4 2 2 3 "1"
                  5 "1" 2 "1" "1"
                  "1" "1" "1" 2 3]]

       ; Act
       (let [layers (form-binary-layers-for-neighbour-cells land-use-map
                                                            origin
                                                            5)]

         ; Assert
         (is (= (count layers) 9))

         ; left-upper-neighbour
         (is (= (nth layers 0)
                [nil nil nil nil nil
                 nil 0   0   nil nil
                 nil 1   1   nil nil
                 nil 0   1   nil nil
                 nil 0   1   1   nil
                 nil 0   nil 0   nil
                 nil nil nil nil nil]))

         ; center-upper-neighbour
         (is (= (nth layers 1)
                [nil nil nil nil nil
                 nil 0   1   nil nil
                 nil 1   0   nil nil
                 nil 1   1   nil nil
                 nil 1   1   1   nil
                 nil 0   nil 0   nil
                 nil nil nil nil nil]))

         ; right-upper-neighbour
         (is (= (nth layers 2)
                [nil nil nil nil nil
                 nil 1   1   nil nil
                 nil 0   0   nil nil
                 nil 1   1   nil nil
                 nil 1   1   1   nil
                 nil 0   nil 1   nil
                 nil nil nil nil nil]))

         ; left-middle-neighbour
         (is (= (nth layers 3)
                [nil nil nil nil nil
                 nil 1   1   nil nil
                 nil 0   1   nil nil
                 nil 0   1   nil nil
                 nil 0   0   0   nil
                 nil 0   nil 0   nil
                 nil nil nil nil nil]))

         ; center-middle-neighbour
         (is (= (nth layers 4)
                [nil nil nil nil nil
                 nil 1   0   nil nil
                 nil 1   1   nil nil
                 nil 1   1   nil nil
                 nil 0   0   0   nil
                 nil 1   nil 1   nil
                 nil nil nil nil nil]))

         ; right-middle-neighbour
         (is (= (nth layers 5)
                [nil nil nil nil nil
                 nil 0   0   nil nil
                 nil 1   1   nil nil
                 nil 1   1   nil nil
                 nil 0   0   1   nil
                 nil 0   nil 1   nil
                 nil nil nil nil nil]))

         ; left-lower-neighbour
         (is (= (nth layers 6)
                [nil nil nil nil nil
                 nil 0   1   nil nil
                 nil 0   1   nil nil
                 nil 0   0   nil nil
                 nil 0   1   0   nil
                 nil 1   nil 1   nil
                 nil nil nil nil nil]))

         ; center-lower-neighbour
         (is (= (nth layers 7)
                [nil nil nil nil nil
                 nil 1   1   nil nil
                 nil 1   1   nil nil
                 nil 0   0   nil nil
                 nil 1   0   1   nil
                 nil 1   nil 0   nil
                 nil nil nil nil nil]))

         ; right-lower-neighbour
         (is (= (nth layers 8)
                [nil nil nil nil nil
                 nil 1   1   nil nil
                 nil 1   1   nil nil
                 nil 0   0   nil nil
                 nil 0   1   1   nil
                 nil 1   nil 0   nil
                 nil nil nil nil nil]))))))

(deftest apply-log-operation
  (testing "Log operation should be applied on map values."
    (
     ; Act
     let [log-map (transform [nil nil nil
                              nil "1" "2"
                              nil "1" "1"
                              nil "1" "1.4"]
                             []
                             :log)]

       ; Assert
       (is (= log-map
              [nil nil nil
               nil 0.0 (Math/log 2)
               nil 0.0 0.0
               nil 0.0 (Math/log 1.4)])))))

(deftest return-self-operation
  (testing "The same map should be returned if no transformation is specified."
    (
     ; Act
     let [self (transform [nil "1"
                           "2" nil
                           "2.3" "5.6"]
                          []
                          :identity)]

       ; Assert
       (is (= self
              [nil "1"
               "2" nil
               "2.3" "5.6"])))))

(deftest if-operation-id-is-unknown-return-map
  (testing "The same map should be returned if operation is unknown."
    (
     ; Act
     let [self (transform [nil "1"
                           "2" nil
                           "2.3" "5.6"]
                          []
                          :unknown-operation)]

       ; Assert
       (is (= self
              [nil "1"
               "2" nil
               "2.3" "5.6"])))))

(deftest apply-unit-rescaling
  (testing "Scale to [0,1] range operation should be applied on map values,
    where value range is set from the original map."
    (
     ; Act
     let [unit-map (transform [nil nil nil
                               nil "1" "2"
                               nil "1" "1"
                               nil "1" "2.8"]
                              [nil nil nil
                               nil "1" "2"
                               nil "1" "1"
                               "3" "1" "2.8"]
                              :unit-rescaling)]

       ; Assert
       (is (= (nth unit-map 0)
              (nth unit-map 1)
              (nth unit-map 2)
              (nth unit-map 3)
              nil))
       (is (double= (nth unit-map 4)
                    0.0))
       (is (double= (nth unit-map 5)
                    0.5))
       (is (= (nth unit-map 6)
              nil))
       (is (double= (nth unit-map 7)
                    0.0))
       (is (double= (nth unit-map 8)
                    0.0))
       (is (= (nth unit-map 9)
              nil))
       (is (double= (nth unit-map 10)
                    0.0))
       (is (double= (nth unit-map 11)
                    0.9)))))


;;; test grouping

(deftest transform-test
  (testing "Integration testing - map values should be transformed with the specified oepration:\n"
    (apply-log-operation)
    (return-self-operation)
    (if-operation-id-is-unknown-return-map)
    (apply-unit-rescaling)))

;;; tests in the namespace

(defn test-ns-hook
  "summary: Explicit definition of tests in the namespace.
   returns: Explicit definition of tests in the namespace."
  []
  (form-binary-layers-for-neighbour-cells-test)
  (transform-test))

