(ns regression-tests.bootstrap-test
  (:require [clojure.test :refer :all]
            [regression-tests.bootstrap :refer :all]
            [utilities.floating-point-comparison :refer :all]))

;;; tests

(deftest regression-stats-percentiles-test
  (testing "Confidence intervals for the statistics in the regression
    model should be estimated."
    (
     ; Act
     let [intervals
          (regression-stats-percentiles [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9]
                                        [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5]
                                         [3 3.1 5] [3.5 3 5.5][4 3 6] [3.8 2.5 6.3]
                                         [3.9 2.7 6.5] [4.2 3.4 6]]
                                        [[0 1 2 0 4 5 6 7 8 6]
                                         [0 1 5 3 5 5 6 7 0 9]
                                         [0 1 2 3 5 5 6 7 8 9]
                                         [0 1 2 3 4 5 8 7 8 9]])]

       ; Assert
       (is (double= (nth (get intervals :coefficients)
                         0)
                    [0.509545459	1.008406331 0.810852898]))
       (is (double= (nth (get intervals :coefficients)
                         1)
                    [-0.63647335	-0.117615299 -0.410165391]))
       (is (double= (nth (get intervals :coefficients)
                         2)
                    [-0.01428938	0.262770518 0.098157801]))
       (is (double= (nth (get intervals :coefficients)
                         3)
                    [-0.387294619	0.736617075 0.238792067]))

       (is (double= (get intervals :r-squared)
                    [0.940166427126323 0.957363826247218 0.948597281]))
       (is (double= (get intervals :mse)
                    [0.051196497 0.074843922 0.062513532])))))

(deftest get-quartiles-general
  (testing "Confidence intervals for the accuracy of the provided observed values
    should be estimated for all sample."
    (
     ; Arrange
     let [values {:y [1 2 1.1 1.4 3 2.1 2.4 1.3 1.5 3.4 0.1 0 1]
                  :urban [0 1 1 2 3 4 5 5 5 6 7 8 9]
                  :x [[1 1 1] [1 1 1] [1.1 1.3 1.1]
                      [2 1 3] [3 1 2] [1 1 2]
                      [3 1 3] [1 0 1] [2 1 0]
                      [3 4 1] [3 1 1] [1 1 1]
                      [1 1 1]]}]
       ; Act
       (let [accuracy (accuracy-percentiles [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9]
                                            [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5]
                                             [3 3.1 5] [3.5 3 5.5][4 3 6] [3.8 2.5 6.3]
                                             [3.9 2.7 6.5] [4.2 3.4 6]]
                                            values
                                            [[0 1 2 0 4 5 6 7 8 6]
                                             [0 1 5 3 5 5 6 7 0 9]
                                             [0 1 2 3 5 5 6 7 8 9]
                                             [0 1 2 3 4 5 8 7 8 9]])]

         ; Assert
         (is (double= (get-in accuracy [:p-value ["all" :p-min]])
                      [0.07831668 0.128893572 0.098244308]))
         (is (double= (get-in accuracy [:p-value ["all" :p-25-percent]])
                      [0.097269634 0.732593941 0.306091468]))
         (is (double= (get-in accuracy [:p-value ["all" :p-50-percent]])
                      [0.391023598 0.914978642 0.569725236]))
         (is (double= (get-in accuracy [:p-value ["all" :p-75-percent]])
                      [1.020028703 1.450732505 1.226903439]))
         (is (double= (get-in accuracy [:p-value ["all" :p-max]])
                      [2.215105965 3.011073338 2.525272318]))))))

(deftest get-quartiles-by-group
  (testing "Confidence intervals for the accuracy of the provided observed values
    should be estimated in the selected group."
    (
     ; Arrange
     let [values {:y [1 2 1.1 1.4 3 2.1 2.4 1.3 1.5 3.4 0.1 0 1]
                  :urban [0 1 1 2 3 4 5 5 5 6 7 8 9]
                  :x [[1 1 1] [1 1 1] [1.1 1.3 1.1]
                      [2 1 3] [3 1 2] [1 1 2]
                      [3 1 3] [1 0 1] [2 1 0]
                      [3 4 1] [3 1 1] [1 1 1]
                      [1 1 1]]}]
       ; Act
       (let [accuracy (accuracy-percentiles [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9]
                                            [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5]
                                             [3 3.1 5] [3.5 3 5.5] [4 3 6] [3.8 2.5 6.3]
                                             [3.9 2.7 6.5] [4.2 3.4 6]]
                                            values
                                            [[0 1 2 0 4 5 6 7 8 6]
                                             [0 1 5 3 5 5 6 7 0 9]
                                             [0 1 2 3 5 5 6 7 8 9]
                                             [0 1 2 3 4 5 8 7 8 9]])]

         ; Assert
         (is (double= (get-in accuracy [:p-value [0 :p-min]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [0 :p-25-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [0 :p-50-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [0 :p-75-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [0 :p-max]])
                      [0.094260676 0.732593941 0.300066896]))

         (is (double= (get-in accuracy [:p-value [1 :p-min]])
                      [0.097269634 0.790646933 0.394511173]))
         (is (double= (get-in accuracy [:p-value [1 :p-25-percent]])
                      [0.097269634 0.790646933 0.394511173]))
         (is (double= (get-in accuracy [:p-value [1 :p-50-percent]])
                      [0.097269634 0.790646933 0.394511173]))
         (is (double= (get-in accuracy [:p-value [1 :p-75-percent]])
                      [0.905739324 1.732593941 1.262362625]))
         (is (double= (get-in accuracy [:p-value [1 :p-max]])
                      [0.905739324 1.732593941 1.262362625]))

         (is (double= (get-in accuracy [:p-value [2 :p-min]])
                      [0.097507446 0.674088248 0.383808854]))
         (is (double= (get-in accuracy [:p-value [2 :p-25-percent]])
                      [0.097507446 0.674088248 0.383808854]))
         (is (double= (get-in accuracy [:p-value [2 :p-50-percent]])
                      [0.097507446 0.674088248 0.383808854]))
         (is (double= (get-in accuracy [:p-value [2 :p-75-percent]])
                      [0.097507446 0.674088248 0.383808854]))
         (is (double= (get-in accuracy [:p-value [2 :p-max]])
                      [0.097507446 0.674088248 0.383808854]))

         (is (double= (get-in accuracy [:p-value [3 :p-min]])
                      [0.096783958 1.450732505 0.581212611]))
         (is (double= (get-in accuracy [:p-value [3 :p-25-percent]])
                      [0.096783958 1.450732505 0.581212611]))
         (is (double= (get-in accuracy [:p-value [3 :p-50-percent]])
                      [0.096783958 1.450732505 0.581212611]))
         (is (double= (get-in accuracy [:p-value [3 :p-75-percent]])
                      [0.096783958 1.450732505 0.581212611]))
         (is (double= (get-in accuracy [:p-value [3 :p-max]])
                      [0.096783958 1.450732505 0.581212611]))

         (is (double= (get-in accuracy [:p-value [4 :p-min]])
                      [1.020028703 1.569823423 1.264204824]))
         (is (double= (get-in accuracy [:p-value [4 :p-25-percent]])
                      [1.020028703 1.569823423 1.264204824]))
         (is (double= (get-in accuracy [:p-value [4 :p-50-percent]])
                      [1.020028703 1.569823423 1.264204824]))
         (is (double= (get-in accuracy [:p-value [4 :p-75-percent]])
                      [1.020028703 1.569823423 1.264204824]))
         (is (double= (get-in accuracy [:p-value [4 :p-max]])
                      [1.020028703 1.569823423 1.264204824]))

         (is (double= (get-in accuracy [:p-value [5 :p-min]])
                      [0.07831668 0.587961987 0.263629886]))
         (is (double= (get-in accuracy [:p-value [5 :p-25-percent]])
                      [0.07831668 0.587961987 0.263629886]))
         (is (double= (get-in accuracy [:p-value [5 :p-50-percent]])
                      [0.107353439 0.914978642 0.41258071]))
         (is (double= (get-in accuracy [:p-value [5 :p-75-percent]])
                      [0.199623425 0.985818999 0.509609365]))
         (is (double= (get-in accuracy [:p-value [5 :p-max]])
                      [0.199623425 0.985818999 0.509609365]))

         (is (double= (get-in accuracy [:p-value [6 :p-min]])
                      [2.194299661 2.466348919 2.271153002]))
         (is (double= (get-in accuracy [:p-value [6 :p-25-percent]])
                      [2.194299661 2.466348919 2.271153002]))
         (is (double= (get-in accuracy [:p-value [6 :p-50-percent]])
                      [2.194299661 2.466348919 2.271153002]))
         (is (double= (get-in accuracy [:p-value [6 :p-75-percent]])
                      [2.194299661 2.466348919 2.271153002]))
         (is (double= (get-in accuracy [:p-value [6 :p-max]])
                      [2.194299661 2.466348919 2.271153002]))

         (is (double= (get-in accuracy [:p-value [7 :p-min]])
                      [1.186496978 3.011073338 2.259343171]))
         (is (double= (get-in accuracy [:p-value [7 :p-25-percent]])
                      [1.186496978 3.011073338 2.259343171]))
         (is (double= (get-in accuracy [:p-value [7 :p-50-percent]])
                      [1.186496978 3.011073338 2.259343171]))
         (is (double= (get-in accuracy [:p-value [7 :p-75-percent]])
                      [1.186496978 3.011073338 2.259343171]))
         (is (double= (get-in accuracy [:p-value [7 :p-max]])
                      [1.186496978 3.011073338 2.259343171]))

         (is (double= (get-in accuracy [:p-value [8 :p-min]])
                      [0.267406059 1.094260676 0.737637375]))
         (is (double= (get-in accuracy [:p-value [8 :p-25-percent]])
                      [0.267406059 1.094260676 0.737637375]))
         (is (double= (get-in accuracy [:p-value [8 :p-50-percent]])
                      [0.267406059 1.094260676 0.737637375]))
         (is (double= (get-in accuracy [:p-value [8 :p-75-percent]])
                      [0.267406059 1.094260676 0.737637375]))
         (is (double= (get-in accuracy [:p-value [8 :p-max]])
                      [0.267406059 1.094260676 0.737637375]))

         (is (double= (get-in accuracy [:p-value [9 :p-min]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [9 :p-25-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [9 :p-50-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [9 :p-75-percent]])
                      [0.094260676 0.732593941 0.300066896]))
         (is (double= (get-in accuracy [:p-value [9 :p-max]])
                      [0.094260676 0.732593941 0.300066896]))))))

(deftest get-bootstrap-sample
  (testing "Bootstrap sample for the quartiles should be returned."
    (
     ; Arrange
     let [values {:y [1 2 1.1 1.4 3 2.1 2.4 1.3 1.5 3.4 0.1 0 1]
                  :urban [0 1 1 2 3 4 5 5 5 6 7 8 9]
                  :x [[1 1 1] [1 1 1] [1.1 1.3 1.1]
                      [2 1 3] [3 1 2] [1 1 2]
                      [3 1 3] [1 0 1] [2 1 0]
                      [3 4 1] [3 1 1] [1 1 1]
                      [1 1 1]]}]
       ; Act
       (let [accuracy (accuracy-percentiles [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9]
                                            [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5]
                                             [3 3.1 5] [3.5 3 5.5][4 3 6] [3.8 2.5 6.3]
                                             [3.9 2.7 6.5] [4.2 3.4 6]]
                                            values
                                            [[0 1 2 0 4 5 6 7 8 6]
                                             [0 1 5 3 5 5 6 7 0 9]
                                             [0 1 2 3 5 5 6 7 8 9]
                                             [0 1 2 3 4 5 8 7 8 9]])]

         ; Assert
         (is (double= (get-in accuracy [:sample :p-min])
                      [0.097507446 0.094260676 0.128893572 0.092243163 0.07831668]))
         (is (double= (get-in accuracy [:sample :p-25-percent])
                      [0.732593941	0.097269634	0.158485872	0.272682594 0.2694253]))
         (is (double= (get-in accuracy [:sample :p-50-percent])
                      [0.914978642	0.674088248	0.474098121	0.3944375732 0.391023598]))
         (is (double= (get-in accuracy [:sample :p-75-percent])
                      [1.450732505	1.020028703	1.131371968	1.264115731 1.268268287]))
         (is (double= (get-in accuracy [:sample :p-max])
                      [2.466348919	3.011073338	2.708930135	2.224903231 2.215105965]))))))

(deftest residuals-independence-test
  (testing "Stat tests should be bootstrapped to check independence in residuals."
    (
     ; Act
     let [result
          (residuals-independence [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9]
                                  [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5]
                                   [3 3.1 5] [3.5 3 5.5][4 3 6] [3.8 2.5 6.3]
                                   [3.9 2.7 6.5] [4.2 3.4 6]]
                                  {0 [1 5]
                                   1 [0]
                                   3 [9]
                                   5 [0 7 8 9]
                                   7 [5]
                                   8 [5]
                                   9 [3 5]}
                                  [[0 1 2 0 4 5 6 7 8 6]
                                   [0 1 5 3 5 5 6 7 0 9]
                                   [0 1 2 3 5 5 6 7 8 9]
                                   [0 1 2 3 4 5 8 7 8 9]])]

       ; Assert
       (is (double= (get-in result [:tests :morans-i-test])
                    [-0.497197740828216 -0.184984968964908
                     -0.340847284475637 2/5]))

       (is (double= (get-in result [:tests :geary-c-test])
                    [0.953671521804377 2.35284987913462
                     1.51454929694469 4/5]))

       (is (double= (get-in result [:samples :morans-i-stat])
                    [-0.45324978913939	-0.497197740828216 -0.184984968964908
                     -0.298642827973139 -0.27016109547253]))

       (is (double= (get-in result [:samples :geary-c-stat])
                    [2.35284987913462 1.38087785015786 0.953671521804377
                     1.45717331902477 1.42817391460185])))))


;;; test grouping

(deftest accuracy-percentiles-test
  (testing "Confidence intervals for quartiles:\n"
    (get-quartiles-general)
    (get-quartiles-by-group)
    (get-bootstrap-sample)))


;;; tests in the namespace

(defn test-ns-hook
  "summary: Explicit definition of tests in the namespace.
   returns: Explicit definition of tests in the namespace."
  []
  (accuracy-percentiles-test)
  (regression-stats-percentiles-test)
  (residuals-independence-test))
