(ns regression-tests.csv-test
  (:require [clojure.test :refer :all]
            [regression-tests.csv :refer :all]
            [utilities.floating-point-comparison :refer :all]))


;;; tests

(deftest run-permutation-tests-test
  (testing "Permutation tests should be executed."
    (
     ; Act
     let [content
          (run-permutation-tests "test/regression_tests/sample.csv"
                                 (fn[indexes]
                                   (->> (map #(map (partial nth indexes)
                                                   %&)
                                             [0 2 1 8 3 4 5 6 9 7]
                                             [5 0 1 3 2 4 8 6 7 9]
                                             [0 2 3 5 4 1 6 7 8 9])
                                        (apply mapv vector))))]

       ; Assert
       (is (= (get-in content [:tests "permutation_tests.csv"])
              '(["test" "p-value" "lower-bound-ci" "upper-bound-ci"]
                ["overall-test-r2" "0.250000" "-0.174352" "0.674352"]
                ["x1-test-t-stat" "0.000000" "0.000000" "0.000000"]
                ["x2-test-t-stat" "0.000000" "0.000000" "0.000000"]
                ["x3-test-t-stat" "0.250000" "-0.174352" "0.674352"])))

       (is (= (get-in content [:samples "permutation_r2_sample.csv"])
              '(["value"]
                ["0.946715"]
                ["0.748369"]
                ["0.797480"]
                ["0.699669"]))))))

(deftest bootstrap-regression-test
  (testing "Estimates from the regression model should be
    obtained by bootstrapping."
    (
     ; Act
     let [content
          (bootstrap-regression "test/regression_tests/sample.csv"
                                (fn[indexes]
                                  (->> (map #(map (partial nth indexes)
                                                  %&)
                                            [0 1 2 0 4 5 6 7 8 6]
                                            [0 1 5 3 5 5 6 7 0 9]
                                            [0 1 2 3 5 5 6 7 8 9]
                                            [0 1 2 3 4 5 8 7 8 9])
                                       (apply mapv vector))))]

       ; Assert
       (is (= (get content "regression-stat-bootstrap.csv")
              '(["statistics" "95-percent-ci-1" "95-percent-ci-2" "mean"]
                ["x1" "0.509545"	"1.008406" "0.810853"]
                ["x2" "-0.636473"	"-0.117615" "-0.410165"]
                ["x3" "-0.014289"	"0.262771" "0.098158"]
                ["intercept" "-0.387295"	"0.736617" "0.238792"]
                ["r-squared" "0.940166" "0.957364" "0.948597"]
                ["mse" "0.051196" "0.074844" "0.062514"]))))))

(deftest bootstrap-accuracy-test
  (testing "Bootstrapping of the accuracy estimates should be performed."
    (
     ; Act
     let [content
          (bootstrap-accuracy "test/regression_tests/sample.csv"
                              "test/regression_tests/values.csv"
                              (fn[indexes]
                                (->> (map #(map (partial nth indexes)
                                                %&)
                                          [0 1 2 0 4 5 6 7 8 6]
                                          [0 1 5 3 5 5 6 7 0 9]
                                          [0 1 2 3 5 5 6 7 8 9]
                                          [0 1 2 3 4 5 8 7 8 9])
                                     (apply mapv vector))))]

       ; Assert
       (is (= (get-in content [:ci "accuracy-bootstrap.csv"])
              '(["id" "group-id" "95-percent-ci-1" "95-percent-ci-2" "mean"]
                ["p-25-percent-0" "0" "0.094261" "0.732594" "0.300067"]
                ["p-25-percent-1" "1" "0.097270" "0.790647" "0.394511"]
                ["p-25-percent-2" "2" "0.097507" "0.674088" "0.383809"]
                ["p-25-percent-3" "3" "0.096784" "1.450733" "0.581213"]
                ["p-25-percent-4" "4" "1.020029" "1.569823" "1.264205"]
                ["p-25-percent-5" "5" "0.078317" "0.587962" "0.263630"]
                ["p-25-percent-6" "6" "2.194300" "2.466349" "2.271153"]
                ["p-25-percent-7" "7" "1.186497" "3.011073" "2.259343"]
                ["p-25-percent-8" "8" "0.267406" "1.094261" "0.737637"]
                ["p-25-percent-9" "9" "0.094261" "0.732594" "0.300067"]
                ["p-25-percent-all" "all" "0.097270" "0.732594" "0.306091"]
                ["p-50-percent-0" "0" "0.094261" "0.732594" "0.300067"]
                ["p-50-percent-1" "1" "0.097270" "0.790647" "0.394511"]
                ["p-50-percent-2" "2" "0.097507" "0.674088" "0.383809"]
                ["p-50-percent-3" "3" "0.096784" "1.450733" "0.581213"]
                ["p-50-percent-4" "4" "1.020029" "1.569823" "1.264205"]
                ["p-50-percent-5" "5" "0.107353" "0.914979" "0.412581"]
                ["p-50-percent-6" "6" "2.194300" "2.466349" "2.271153"]
                ["p-50-percent-7" "7" "1.186497" "3.011073" "2.259343"]
                ["p-50-percent-8" "8" "0.267406" "1.094261" "0.737637"]
                ["p-50-percent-9" "9" "0.094261" "0.732594" "0.300067"]
                ["p-50-percent-all" "all" "0.391024" "0.914979" "0.569725"]
                ["p-75-percent-0" "0" "0.094261" "0.732594" "0.300067"]
                ["p-75-percent-1" "1" "0.905739" "1.732594" "1.262363"]
                ["p-75-percent-2" "2" "0.097507" "0.674088" "0.383809"]
                ["p-75-percent-3" "3" "0.096784" "1.450733" "0.581213"]
                ["p-75-percent-4" "4" "1.020029" "1.569823" "1.264205"]
                ["p-75-percent-5" "5" "0.199623" "0.985819" "0.509609"]
                ["p-75-percent-6" "6" "2.194300" "2.466349" "2.271153"]
                ["p-75-percent-7" "7" "1.186497" "3.011073" "2.259343"]
                ["p-75-percent-8" "8" "0.267406" "1.094261" "0.737637"]
                ["p-75-percent-9" "9" "0.094261" "0.732594" "0.300067"]
                ["p-75-percent-all" "all" "1.020029" "1.450733" "1.226903"]
                ["p-max-0" "0" "0.094261" "0.732594" "0.300067"]
                ["p-max-1" "1" "0.905739" "1.732594" "1.262363"]
                ["p-max-2" "2" "0.097507" "0.674088" "0.383809"]
                ["p-max-3" "3" "0.096784" "1.450733" "0.581213"]
                ["p-max-4" "4" "1.020029" "1.569823" "1.264205"]
                ["p-max-5" "5" "0.199623" "0.985819" "0.509609"]
                ["p-max-6" "6" "2.194300" "2.466349" "2.271153"]
                ["p-max-7" "7" "1.186497" "3.011073" "2.259343"]
                ["p-max-8" "8" "0.267406" "1.094261" "0.737637"]
                ["p-max-9" "9" "0.094261" "0.732594" "0.300067"]
                ["p-max-all" "all" "2.215106" "3.011073" "2.525272"]
                ["p-min-0" "0" "0.094261" "0.732594" "0.300067"]
                ["p-min-1" "1" "0.097270" "0.790647" "0.394511"]
                ["p-min-2" "2" "0.097507" "0.674088" "0.383809"]
                ["p-min-3" "3" "0.096784" "1.450733" "0.581213"]
                ["p-min-4" "4" "1.020029" "1.569823" "1.264205"]
                ["p-min-5" "5" "0.078317" "0.587962" "0.263630"]
                ["p-min-6" "6" "2.194300" "2.466349" "2.271153"]
                ["p-min-7" "7" "1.186497" "3.011073" "2.259343"]
                ["p-min-8" "8" "0.267406" "1.094261" "0.737637"]
                ["p-min-9" "9" "0.094261" "0.732594" "0.300067"]
                ["p-min-all" "all" "0.078317" "0.128894" "0.098244"])))

       (is (= (get-in content [:samples "accuracy-sample.csv"])
              '(["min" "quartile-1" "quartile-2" "quartile-3" "max"]
                ["0.097507" "0.732594" "0.914979" "1.450733" "2.466349"]
                ["0.094261" "0.097270" "0.674088" "1.020029" "3.011073"]
                ["0.128894" "0.158486" "0.474098" "1.131372" "2.708930"]
                ["0.092243" "0.272683" "0.394438" "1.264116" "2.224903"]
                ["0.078317" "0.269425" "0.391024" "1.268268" "2.215106"]))))))

(deftest bootstrap-independence-tests-test
  (testing "Independence tests should be executed by bootstrapping."
    (
     ; Act
     let [content
          (bootstrap-independence-tests "test/regression_tests/sample.csv"
                                        "test/regression_tests/neighbours.csv"
                                        (fn[indexes]
                                          (->> (map #(map (partial nth indexes)
                                                          %&)
                                                    [0 1 2 0 4 5 6 7 8 6]
                                                    [0 1 5 3 5 5 6 7 0 9]
                                                    [0 1 2 3 5 5 6 7 8 9]
                                                    [0 1 2 3 4 5 8 7 8 9])
                                               (apply mapv vector))))]

       ; Assert
       (is (= (get content "independence-tests-bootstrap.csv")
              '(["statistics" "95-percent-ci-1" "95-percent-ci-2" "mean" "p-value"]
                ["morans-i-test" "-0.497198" "-0.184985" "-0.340847" "0.400000"]
                ["geary-c-test" "0.953672" "2.352850" "1.514549" "0.800000"])))

       (is (= (get content "morans-i-test-sample.csv")
              '(["value"]
                ["-0.453250"]
                ["-0.497198"]
                ["-0.184985"]
                ["-0.298643"]
                ["-0.270161"])))

       (is (= (get content "geary-c-test-sample.csv")
              '(["value"]
                ["2.352850"]
                ["1.380878"]
                ["0.953672"]
                ["1.457173"]
                ["1.428174"]))))))
